
/*C*

________________________________________________________________

        matlab_region
        $Id: matlab_region.h,v 1.5 2005/08/12 08:39:54 svein Exp $
        Copyright 1990-2003, Blab, UiO
        Image processing lab, Department of Informatics
        University of Oslo
        E-mail: blab@ifi.uio.no
________________________________________________________________
  
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose and without fee is hereby granted, 
  provided that this copyright notice appear in all copies and that 
  both that copyright notice and this permission notice appear in supporting
  documentation and that the name of B-lab, Department of Informatics or
  University of Oslo not be used in advertising or publicity pertaining 
  to distribution of the software without specific, written prior permission.

  B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL B-LAB
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 

*/

#ifndef _MATLAB_REGION_H_
#define _MATLAB_REGION_H_

_XITE_CPLUSPLUS_BEGIN

#include <matrix.h>

#define BEHAVIOUR_FIELD_HEADER  "header"
#define BEHAVIOUR_FIELD_NC      "nc"
#define BEHAVIOUR_FIELD_SORT    "sort"
#define BEHAVIOUR_FIELD_SPLIT   "split"
#define BEHAVIOUR_FIELD_CURSORX "cursorx"
#define BEHAVIOUR_FIELD_CURSORY "cursory"
#define BEHAVIOUR_FIELD_BG      "bg"
#define BEHAVIOUR_FIELD_IFLINE  "ifline"
#define BEHAVIOUR_FIELD_OUTFILE "outfile"

static const char *behaviour_field_names_s[] = {
  BEHAVIOUR_FIELD_HEADER,
  BEHAVIOUR_FIELD_NC,
  BEHAVIOUR_FIELD_SORT,
  BEHAVIOUR_FIELD_SPLIT,
  BEHAVIOUR_FIELD_CURSORX,
  BEHAVIOUR_FIELD_CURSORY,
  BEHAVIOUR_FIELD_BG,
  BEHAVIOUR_FIELD_IFLINE,
  BEHAVIOUR_FIELD_OUTFILE,
};
static mxClassID behaviour_field_classes_s[] = {
  mxINT32_CLASS, /* header */
  mxINT32_CLASS, /* nc */
  mxINT32_CLASS, /* sort */
  mxINT32_CLASS, /* split */
  mxINT32_CLASS, /* cursorx */
  mxINT32_CLASS, /* cursory */
  mxINT32_CLASS, /* bg */
  mxCHAR_CLASS,  /* ifline */
  mxCHAR_CLASS,  /* outfile */
};

#define PROPERTIES_FIELD_OBJECTNR  "objectnr"
#define PROPERTIES_FIELD_REGIONNR  "regionnr"
#define PROPERTIES_FIELD_AREA      "area"
#define PROPERTIES_FIELD_PERIMETER "perimeter"
#define PROPERTIES_FIELD_XMIN      "xmin"
#define PROPERTIES_FIELD_YMIN      "ymin"
#define PROPERTIES_FIELD_XMAX      "xmax"
#define PROPERTIES_FIELD_YMAX      "ymax"
#define PROPERTIES_FIELD_HEIGHT    "height"
#define PROPERTIES_FIELD_WIDTH     "width"
#define PROPERTIES_FIELD_XPOINT    "xpoint"
#define PROPERTIES_FIELD_YPOINT    "ypoint"
#define PROPERTIES_FIELD_MINPIX    "minpix"
#define PROPERTIES_FIELD_MAXPIX    "maxpix"
#define PROPERTIES_FIELD_XMEAN     "xmean"
#define PROPERTIES_FIELD_YMEAN     "ymean"
#define PROPERTIES_FIELD_MOMENT    "moment"
#define PROPERTIES_FIELD_MOMENTP   "momentp"
#define PROPERTIES_FIELD_MOMENTQ   "momentq"
#define PROPERTIES_FIELD_CMOMENT   "cmoment"
#define PROPERTIES_FIELD_CMOMENTP  "cmomentp"
#define PROPERTIES_FIELD_CMOMENTQ  "cmomentq"
#define PROPERTIES_FIELD_IMOMENT   "imoment"
#define PROPERTIES_FIELD_VOLUME    "volume"
#define PROPERTIES_FIELD_MEAN      "mean"
#define PROPERTIES_FIELD_MEDIAN    "median"
#define PROPERTIES_FIELD_STDEV     "stdev"

static const char *properties_field_names_s[] = {
  PROPERTIES_FIELD_OBJECTNR,
  PROPERTIES_FIELD_REGIONNR,
  PROPERTIES_FIELD_AREA,
  PROPERTIES_FIELD_PERIMETER,
  PROPERTIES_FIELD_XMIN,
  PROPERTIES_FIELD_YMIN,
  PROPERTIES_FIELD_XMAX,
  PROPERTIES_FIELD_YMAX,
  PROPERTIES_FIELD_HEIGHT,
  PROPERTIES_FIELD_WIDTH,
  PROPERTIES_FIELD_XPOINT,
  PROPERTIES_FIELD_YPOINT,
  PROPERTIES_FIELD_MINPIX,
  PROPERTIES_FIELD_MAXPIX,
  PROPERTIES_FIELD_XMEAN,
  PROPERTIES_FIELD_YMEAN,
  PROPERTIES_FIELD_MOMENT,
  PROPERTIES_FIELD_MOMENTP,
  PROPERTIES_FIELD_MOMENTQ,
  PROPERTIES_FIELD_CMOMENT,
  PROPERTIES_FIELD_CMOMENTP,
  PROPERTIES_FIELD_CMOMENTQ,
  PROPERTIES_FIELD_IMOMENT,
  PROPERTIES_FIELD_VOLUME,
  PROPERTIES_FIELD_MEAN,
  PROPERTIES_FIELD_MEDIAN,
  PROPERTIES_FIELD_STDEV,
};

static int properties_field_classes_s[] = {
  mxINT32_CLASS,  /* objectnr */
  mxINT32_CLASS,  /* regionnr */
  mxINT32_CLASS,  /* area */
  mxINT32_CLASS,  /* perimeter */
  mxINT32_CLASS,  /* xmin */
  mxINT32_CLASS,  /* ymin */
  mxINT32_CLASS,  /* xmax */
  mxINT32_CLASS,  /* ymax */
  mxINT32_CLASS,  /* height */
  mxINT32_CLASS,  /* width */
  mxINT32_CLASS,  /* xpoint */
  mxINT32_CLASS,  /* ypoint */
  mxINT32_CLASS,  /* minpix */
  mxINT32_CLASS,  /* maxpix */
  mxDOUBLE_CLASS, /* xmean */
  mxDOUBLE_CLASS, /* ymean */
  mxDOUBLE_CLASS, /* moment */
  mxINT32_CLASS,  /* momentp */
  mxINT32_CLASS,  /* momentq */
  mxDOUBLE_CLASS, /* cmoment */
  mxINT32_CLASS,  /* cmomentp */
  mxINT32_CLASS,  /* cmomentq */
  mxDOUBLE_CLASS, /* imoment */
  mxINT32_CLASS,  /* volume */
  mxDOUBLE_CLASS, /* mean */
  mxINT32_CLASS,  /* median */
  mxDOUBLE_CLASS, /* stdev */
};

static int num_behaviour_fields_s =
  sizeof(behaviour_field_names_s) / sizeof(char *);
static int num_properties_fields_s =
  sizeof(properties_field_names_s) / sizeof(char *);


_XITE_CPLUSPLUS_END

#endif /* _MATLAB_REGION_H_ */
