
/*C*

________________________________________________________________

        region2matlab
        $Id: region2matlab.c,v 1.8 2005/08/12 08:39:12 svein Exp $
        Copyright 1994, Blab, UiO
        Image processing lab, Department of Informatics
        University of Oslo
        E-mail: blab@ifi.uio.no
________________________________________________________________
  
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose and without fee is hereby granted, 
  provided that this copyright notice appear in all copies and that 
  both that copyright notice and this permission notice appear in supporting
  documentation and that the name of B-lab, Department of Informatics or
  University of Oslo not be used in advertising or publicity pertaining 
  to distribution of the software without specific, written prior permission.

  B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL B-LAB
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 

*/

static char *Id = "$Id: region2matlab.c,v 1.8 2005/08/12 08:39:12 svein Exp $, Blab, UiO";

#include <mat.h>

#include <xite/includes.h>
#include <xite/matlab.h>
#include "matlab_region.h"
#include <xite/message.h>
#include <xite/debug.h>
#include XITE_STRING_H

#ifndef MAIN

#define OK_IND                 0
#define CONVERT_PROPERTIES_IND 1

static char *error_s[] = {
  "OK\n",
  "Failed in converting XITE region_properties struct to MATLAB array.\n",
};

int region_properties2matlab(region_properties *prop, mxArray **array) {
  int field_num;
  mxArray *arr = mxCreateStructMatrix(1, 1, num_properties_fields_s,
				      properties_field_names_s);

  for (field_num = 0; field_num < num_properties_fields_s; field_num++) {
    const char *name      = properties_field_names_s[field_num];
    mxClassID field_class = properties_field_classes_s[field_num];
    mxArray *field;
    fprintf(stderr, "field_num=%d\n", field_num);
    fprintf(stderr, "field_name=%s\n", name);

    if (strcmp(name, PROPERTIES_FIELD_IMOMENT)) {
      field = mxCreateNumericMatrix(1, 1, field_class, mxREAL);
    } else {
      field = mxCreateNumericMatrix(8, 1, field_class, mxREAL);
    }

    if        (!strcmp(name, PROPERTIES_FIELD_OBJECTNR)) {
      ((int *) mxGetData(field))[0] = prop->objectnr;
    } else if (!strcmp(name, PROPERTIES_FIELD_REGIONNR)) {
      ((int *) mxGetData(field))[0] = prop->regionnr;
    } else if (!strcmp(name, PROPERTIES_FIELD_AREA)) {
      ((int *) mxGetData(field))[0] = prop->area;
    } else if (!strcmp(name, PROPERTIES_FIELD_PERIMETER)) {
      ((int *) mxGetData(field))[0] = prop->perimeter;
    } else if (!strcmp(name, PROPERTIES_FIELD_XMIN)) {
      ((int *) mxGetData(field))[0] = prop->xmin;
    } else if (!strcmp(name, PROPERTIES_FIELD_YMIN)) {
      ((int *) mxGetData(field))[0] = prop->ymin;
    } else if (!strcmp(name, PROPERTIES_FIELD_XMAX)) {
      ((int *) mxGetData(field))[0] = prop->xmax;
    } else if (!strcmp(name, PROPERTIES_FIELD_YMAX)) {
      ((int *) mxGetData(field))[0] = prop->ymax;
    } else if (!strcmp(name, PROPERTIES_FIELD_HEIGHT)) {
      ((int *) mxGetData(field))[0] = prop->height;
    } else if (!strcmp(name, PROPERTIES_FIELD_WIDTH)) {
      ((int *) mxGetData(field))[0] = prop->width;
    } else if (!strcmp(name, PROPERTIES_FIELD_XPOINT)) {
      ((int *) mxGetData(field))[0] = prop->xpoint;
    } else if (!strcmp(name, PROPERTIES_FIELD_YPOINT)) {
      ((int *) mxGetData(field))[0] = prop->ypoint;
    } else if (!strcmp(name, PROPERTIES_FIELD_MINPIX)) {
      ((int *) mxGetData(field))[0] = prop->minpix;
    } else if (!strcmp(name, PROPERTIES_FIELD_MAXPIX)) {
      ((int *) mxGetData(field))[0] = prop->maxpix;
    } else if (!strcmp(name, PROPERTIES_FIELD_XMEAN)) {
      ((double *) mxGetData(field))[0] = prop->xmean;
    } else if (!strcmp(name, PROPERTIES_FIELD_YMEAN)) {
      ((double *) mxGetData(field))[0] = prop->ymean;
    } else if (!strcmp(name, PROPERTIES_FIELD_MOMENT)) {
      ((double *) mxGetData(field))[0] = prop->moment->moment;
    } else if (!strcmp(name, PROPERTIES_FIELD_MOMENTP)) {
      ((int *) mxGetData(field))[0] = prop->moment->p;
    } else if (!strcmp(name, PROPERTIES_FIELD_MOMENTQ)) {
      ((int *) mxGetData(field))[0] = prop->moment->q;
    } else if (!strcmp(name, PROPERTIES_FIELD_CMOMENT)) {
      ((double *) mxGetData(field))[0] = prop->cmoment->moment;
    } else if (!strcmp(name, PROPERTIES_FIELD_CMOMENTP)) {
      ((int *) mxGetData(field))[0] = prop->cmoment->p;
    } else if (!strcmp(name, PROPERTIES_FIELD_CMOMENTQ)) {
      ((int *) mxGetData(field))[0] = prop->cmoment->q;
    } else if (!strcmp(name, PROPERTIES_FIELD_IMOMENT)) {
      memcpy((double *) mxGetData(field), prop->imoment, 8 * sizeof(double));
    } else if (!strcmp(name, PROPERTIES_FIELD_VOLUME)) {
      ((int *) mxGetData(field))[0] = prop->volume;
    } else if (!strcmp(name, PROPERTIES_FIELD_MEAN)) {
      ((double *) mxGetData(field))[0] = prop->mean;
    } else if (!strcmp(name, PROPERTIES_FIELD_MEDIAN)) {
      ((int *) mxGetData(field))[0] = prop->median;
    } else if (!strcmp(name, PROPERTIES_FIELD_STDEV)) {
      ((double *) mxGetData(field))[0] = prop->stdev;
    }

    mxSetFieldByNumber(arr, 0, field_num, field);
    
  } /* for */

  *array = arr;

  return(Iok);

} /* region_properties2matlab */

int region_properties_arr2matlab(region_properties **prop, mxArray **array,
				 int num) {
  int ind, status, field_num;
  region_properties *pr;
  mxArray *arr, *field;

  *array = mxCreateStructMatrix(num, 1, num_properties_fields_s,
				properties_field_names_s);

  for (ind = 0; ind < num; ind++) {
    pr = prop[ind];
    status = region_properties2matlab(pr, &arr);
    if (status) {
      return(Error(CONVERT_PROPERTIES_IND, error_s[CONVERT_PROPERTIES_IND]));
    }

    for (field_num = 0; field_num < num_properties_fields_s; field_num++) {
      field = mxGetFieldByNumber(arr, 0, field_num);
      mxSetFieldByNumber(*array, ind, field_num, field);
    }
  }

  return(Iok);
  
} /* region_properties_arr2matlab */

#endif /* not MAIN */
