% XITE Toolbox Version 3.46
%
% XITE is developed by
%
%     Signal Processing and Image Analysis Group
%     Department of Informatics
%     University of Oslo
%     www.mn.uio.no/ifi/english/research/groups/dsb/resources/software/xite
%
% This toolbox interfaces to the XITE functions by passing MATLAB arrays
% to mex files. The arrays are converted in the mex files to corresponding
% XITE C types before an XITE C function is called. Output from the XITE
% C functions are returned to MATLAB as MATLAB arrays.
%
% Naming convention
% -----------------
%
% All XITE function names in this toolbox, except loadBiff and saveBiff, have
% the same names as the corresponding C functions, prefixed with "xite_".
% This is to avoid name collisions with functions from other toolboxes.
%
% MATLAB arrays and XITE images
% -----------------------------
% 
% This toolbox handles conversion between MATLAB arrays and XITE images
% automatically. The numeric types of pixels are preserved in this conversion.
%
% Most XITE functions expect at least one image parameter. These are defined
% by XITEs own C types IBAND, IMAGE and related types. The BAND types
% represent a single matrix. The IMAGE types represent a collection of
% matrices/BANDS.
%
% The correspondence between MATLAB array classes and XITE image/band parameter
% types are
%
%     XITE type                  MATLAB array class
%     ---------------------------------------------
%
%     IBAND                      mxUINT8_CLASS (see below)
%     IUB_BAND                   mxUINT8_CLASS
%     ISB_BAND                   mxINT8_CLASS
%     IUS_BAND                   mxUINT16_CLASS
%     ISS_BAND                   mxINT16_CLASS
%     II_BAND                    mxINT32_CLASS
%     IR_BAND                    mxSINGLE_CLASS
%     IC_BAND                    mxSINGLE_CLASS
%     ID_BAND                    mxDOUBLE_CLASS
%     IDC_BAND                   mxDOUBLE_CLASS
%
%     IBAND is a synonym for IUB_BAND, but often used as a generic type.
%     If the C function header states a parameter type of IBAND, it may well
%     be that it accepts any of the above. The function documentation (in the
%     accompanying m-file) should state what type the function really accepts.
%
%     If the mex-file is handed an array of a class which the XITE C function
%     does not support, an error message will be issued to the MATLAB command
%     window.
%
%     The IMAGE types have names similar to the BAND types: IMAGE,
%     IUB_IMAGE, ISB_IMAGE etc.
%
% MATLAB arrays and XITE non-image data
% -------------------------------------
%
% Non-IMAGE/BAND parameter types expected by the XITE functions will
% accept MATLAB arrays of a corresponding type.
%
% Scalar parameters
% -----------------
%
% For reasons of flexibility, numeric scalar parameters of C type "short",
% "int", "long" and "float" will also accept MATLAB scalars of type
% "double". This way one can use scalar constants without needing to wrap them
% in conversion functions such as int32 and the like:
%
%     [r,arr]=xite_profile(image_array,1,1,0,2,4,1,10);
%
% instead of
%
%     [r,arr]=xite_profile(image_array,1,1,0,2,4,int32(1),int32(10));
%
% Import and export of XITE images
% --------------------------------
%
% XITE images are stored on file in the BIFF format. Images can be imported
% to and exported from MATLAB with the functions
%
%      loadBiff(filename)
%      saveBiff(mat1, mat2, mat3, filename)
%
% Use the MATLAB help command for more information on loadBiff and saveBiff.
%
% Subdirectories
% ---------------
%
% arithmetic
% binarize
% canny
% classification
% combine
% contour
% convolve
% copy
% dither
% extrema
% fft
% fht
% fractal
% geometry
% gradInv
% haar
% histo
% hough
% ideal
% ihs
% kncn
% knn
% median
% minmax
% morph
% negate
% noise
% principal
% pyramid
% resample
% scatter
% segmentation
% sigma
% snn
% statistics
% stdiff
% texture
% thin
% threshold
% thresMl
% zernike
